from .main import ChatGPT

__all__ = ["ChatGPT"]

__version__ = "0.1.7"
__author__ = "Smartwa"
__repo__ = "https://github.com/Simatwa/WebChatGPT"
__info__ = "Reverse Engineering of ChatGPT Web-version."
